

/*****************************************************************************

ScrumPy -- Metabolic Modelling with Python

Copyright Mark Poolman 1995 - 2002

 This file is part of ScrumPy.

    ScrumPy is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    ScrumPy is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with ScrumPy; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


*****************************************************************************/
/********

M.G. Poolman 8/11/99

LowLevelKinMod.h -- C++ class interface to Scampi model API
"low level" wrt python
********/

#ifndef MODEL_H
#define MODEL_H 1

#include <fstream>

#include <ScampiModelBase.h>
#include <Scampi.h>


#include "VarTbl.h"
#include "SSPop.h"


const char Version[]={"VERSION"} ;


const double DefaultSSTol = 1e-6,
			 DefaultSimTol = 1e-6,
			 DefaultTincr = 1.0, 
			 DefaultSSMuteSize = 0.1 ;

const int 	 DefaultSSMaxGens = 50,
			 DefaultNewtIts = 20 ;

const char TimeString[] = {"Time"} ;
  
extern double NottaNum(void) ;

class LowLevelKinMod {

	
  public:
  
	LowLevelKinMod() ;					// raw, !Usable() ;
	~LowLevelKinMod() ;
	
	bool Usable() ;
	  /* pre: true
	    post: Usable <=> model is usable */
		
	void Load(const char *modname, const char *libname) ;
	  /* pre: true
	    post: Attempts to load model from file name. Hints to stderr.
			  User should now test model with Usable(). ErrorMsg() will
			  return a suitable message */     
		
	bool Exists(const char *name) ;
	  /* pre: Usable() ;
	    post: Exists(name) <=> this has a quantity associated with name */
		 
	double GetVal(const char *name) ;
	  /* pre: Exists(name)
	    post: GetVal(name)  == value associated with name */
				
	double *GetVal_p(const char *name) ;
	  /* pre: Exists(name)
	    post: GetVal_p(name)  == pointer tp value associated with name */
		
	
	/* enum Component_id */ int GetId(const char *name) ;
	  /* pre: Exists(Name) ;
	    post: GetId(Name) == id of name */	

				
	
	
	void SetVal(const char *name, double val) ;
		/* pre: Exists(name)
	      post: GetVal(name)  == val */
		
	int SizeVec(enum Component_id comp) ;
		/* pre: Usable() 
		  post: SizeVec(comp) == size of vector associated with comp */

	double *GetVec(enum Component_id comp) ;
		/* pre: Usable()
		  post: GetVec(comp) = (pointer to first element in) **READ ONLY** vector 
		  		associated with comp */
	
	double *CpyVec(enum Component_id comp) ;
		/* pre: Usable()
		  post: returns a new copy of GetVec(comp), which it is the user's 
		  		responsobility do delete */
	
	void NaNVec(enum Component_id comp) ;
		/* pre: Usable()
		  post: All els of vec are  NaN */
		  
	void ZerVec(enum Component_id comp) ;
		/* pre: Usable()
		  post: All els of vec are zero */ 
	
	void SetVec(enum Component_id comp, double *vec) ;
		/* pre: Usable()
		  post: vec is copied into vector associated with comp */
	
	
	int n_Cons(void) ;
		/* pre: Usable() ;
		  post: n_Cons() == number of conserved metabolites */
	
	void UpdateCons(void) ;
		/* pre: Usable()
		  post: Concentrations updated to reflect current csum params */	  
				  		
	char **GetNameVec(enum Component_id comp) ;
		/* pre: Usable()
		  post: returns the **READ ONLY** vector of names in the model assoc w/ Param */
		         
		  
	void SetRescue(void) ;
		/* pre: Usable()
		  post: Rescue concentrations set to current concentrations */
	
	void GetRescue(void) ;
		/* pre: Usable()
		  post: Current concentrations set to rescue concentrations */ 	

	/*** functions affecting ss determinatiopn ***/
	void SetESMute(double m) ; 		// pre: m > 0.0
	void SetESAlt(bool alt) ;		// use altruistic ES for SS, desault false
	void SetESLam(bool lam) ; 		// use Lamarckian ES for SS, default true
	void ResetPop(void) ; 			// return ES Pop to init
		
	/*** functions affecting simulation  ***/
	void SetSimTol(double tol) ;	// pre: tol > 0.0 
	void SetTincr(double tincr) ;	// pre: incr > 0.0
	double GetTincr(void) ;
	void ResetInteg(void) ;			// reset Lsoda
	
	
	/****  report any numerical problems  ****/
	
	bool IsOK(void) ;
		/* pre: true ;
		  post: IsOK() => the last numerical action completed succesfully,
		  		ELSE velocity and concentration values not defined, and
				     should be checked +/- updated */
		 
	bool CSumsOK(void) ;
	   /* pre: true ;
		  post: CSumsOK() => Cureent free met values do not violate 
				conservation relationships */
				  
	const char *ErrorMsg(void) ;
		/* pre: true ;
		  post: ErrorMsg() = Description of most recent error */

	/* SPI_err */ int LastErr(void) ;
		/* pre: true ;
		  post: LastErr() == last error as defined in ScampiConsts.h */

	void ClearErr(void) ;
		/* pre: true ;
		  post: IsOK() even if it isn't !!  */ 
			  
				  
	/*** Simulation and steady-state ****/	  

	void Simulate(int n_points) ;
		/* pre: true
		  post: IsOk() => Succesful simulation to time + tincr, time updated
		  	    ELSE problem with integration, possible hints on sterr */  

	void SimToTime(double tend) ;
		/* pre: tend > this.Time
		  post: simulates to tend */
		  
	double NewtSS(void) ;
		  /* pre: Usable() ;
		  post: Attempts raw Newton to get SS, returns tol achieved */
	  
	void EvSS(void) ;
		/* pre: Usable() ;
		  post: Attempts to use ES with Newton to get SS */

	void Eval(void) ;
		/* pre: true
		  post: model updated to reflect current parameter and concentation vals */		  
		  
		  
	void MinPackFCN(int *N, double *MetVals, double *dMet_dT, int *err) ;
	    /* FCN as required by the minpack HYBRD solver */
		  
  // private:


  	 void Destroy(void) ;   
 	ScampiModel *mod ;
	void *dlib ;
	char *dlibname ;
	VarTbl VarTab ;
	SPI_err err ; 

	double *ConcBuf ;	
		
	SSPop SS_Pop ;
	Organism SS_Adam ;

	bool Alt, Lam ;
		
	double SSTol, SimTol, Tincr, SSMuteSize ;
	int  n_Refs, SimSteps, SSMaxGens, NewtIts ;
	
	double *Time, *RescueConcs ; 
	
	
	void Init(void) ;
	
} ;
  
	



	
#endif
