

/*****************************************************************************

ScrumPy -- Metabolic Modelling with Python

Copyright Mark Poolman 1995 - 2002

 This file is part of ScrumPy.

    ScrumPy is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    ScrumPy is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with ScrumPy; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


*****************************************************************************/

/*
M.G.Poolman - 3/8/06
ScratchPad structure for Newton-Raphson solver  using the meschach library
*/

#include<stdlib.h>
#include<stdio.h>


/* mescach headers */
#include<matrix.h>
#include<matrix2.h>
#include<err.h>

#include "../SS_ScratchPad.h"


#ifndef NULL
#define NULL 0L
#endif


enum WhichOde {OdeLo, OdeMid, OdeHi,n_Ode_ts} ;
typedef enum WhichOde WhichOde_t ;

struct SS_ScratchPad {
	
	MAT *Jac, *InvJac ; /* Jacobian, inverse Jacaobian */
	VEC *Mets, *Odes[n_Ode_ts], *Corrxs, *Perts ; 
	ScampiModel_t mod ;
}  ;



SS_ScratchPad_t NewSS_ScratchPad(ScampiModel_t mod){

  SS_ScratchPad_t rv ;
  int n_Free = mod->n_Free, n ;
  count_errs(0) ; /* meschach/err.h - don't count errors (and thus exit() if >=100 errors) */
  
	rv =  (SS_ScratchPad_t) calloc(1, sizeof(struct SS_ScratchPad)) ;
	
	if(rv != NULL){  /* a bit unlikely, but you never know */
		rv->Jac = m_get(n_Free,n_Free) ;
		rv->InvJac = m_get(n_Free,n_Free) ;
		rv->Mets =  v_get(n_Free) ;
		for (n = 0 ; n < n_Ode_ts; n++)
			rv->Odes[n] = v_get(n_Free) ;
		rv->Corrxs = v_get(n_Free) ;
		rv->Perts =  v_get(n_Free) ;
		rv->mod = mod ;
	}
	if (rv == NULL)
		fprintf(stderr, "\n!!!! calloc fail in hybrd1 NewSS_ScratchPad(%d) !!!!\n", n_Free) ;
	
	return rv ;
} ;  
		




void DelSS_ScratchPad(SS_ScratchPad_t sp){
	return ;
} ;




void ReadOdes(SS_ScratchPad_t sp, WhichOde_t which){

  int n ;
  double *odes = sp->mod->dMet_dT ;
  
	for (n=0 ; n< sp->mod->n_Free ; n++)
		 v_set_val(sp->Odes[which],n,odes[n]) ;
}



void ReadMets(SS_ScratchPad_t sp){

  int n ;
  double  *mets = sp->mod->Met_Vals ;
  
	for (n=0 ; n< sp->mod->n_Free ; n++)
		v_set_val(sp->Mets,n,mets[n]) ;
}


void WriteMets(SS_ScratchPad_t sp){

  int n, n_Free = sp->mod->n_Free ;
  double  *mets = sp->mod->Met_Vals ;
  
	for(n = 0 ; n < n_Free ; n++) 
		mets[n] = v_entry(sp->Mets,n) ;
	
}


void WriteOdes(SS_ScratchPad_t sp, WhichOde_t which){

  int n, n_Free = sp->mod->n_Free;
  double  *odes = sp->mod->dMet_dT ;
  
	for(n = 0 ; n < n_Free ; n++) 
		odes[n] = v_entry(sp->Odes[which],n) ;
	
}
	 
		 

void CalcJac(SS_ScratchPad_t sp){
  
  /* relpert to low ? leads to numerical instability for SS ?? */

  double relpert = 1e-15, *mets = sp->mod->Met_Vals, metval, pert, ddt  ;
  int m,n, n_Free ;
  
  	n_Free = sp->mod->n_Free ;
 
	
	
	for (n=0 ; n < n_Free ; n++){
		metval = mets[n] ;
		pert = relpert * metval ;
		mets[n] -= pert ;
		ModEval(sp->mod) ;
		ReadOdes(sp, OdeLo) ;
		mets[n] += 2.0*pert ;
		ModEval(sp->mod) ;
		ReadOdes(sp, OdeHi) ;
		for (m=0 ; m<n_Free ; m++){
			ddt = (v_entry(sp->Odes[OdeHi],m) - v_entry(sp->Odes[OdeLo],m))/(2.0*pert) ;
			
			m_set_val(sp->Jac,m,n,ddt) ;
		}
		mets[n] = metval ;
	}
}
		


double CalcErr(SS_ScratchPad_t sp){

  double sum = 0.0, ode ;
  int n ;
  
  for (n=0 ; n< sp->mod->n_Free ; n++){
    ode =  sp->mod->dMet_dT[n] ;
    sum += ode*ode ;
  }
  return sqrt(sum) ;
}



double Solve (SS_ScratchPad_t sp,  int Max_Iters, double tol, enum SPI_err *err) {
						
	int niters=0, n ;
	double *mets,*odes, ErrVal ;

	Max_Iters=50 ;

	mets = ModGetVec(sp->mod,Conc) ;
    odes = ModGetVec(sp->mod,dMet) ;
						
	ModEval(sp->mod) ;
	ReadOdes(sp,OdeMid) ;
	ErrVal = CalcErr(sp) ;
	fprintf(stderr, "ini err val %g (%g)\n",ErrVal,tol) ;
	
	catchall( /* meschach error handling macro see err.h */
	   {
	       while( ErrVal >= tol   &&  ++niters<Max_Iters){ /* just yer basic Newton-Raphson - nothing fancy */
	           CalcJac(sp) ;
	           m_inverse(sp->Jac,sp->InvJac) ;
        	   ModEval(sp->mod) ;
        	   ReadOdes(sp,OdeMid) ;
        	   mv_mlt(sp->InvJac,sp->Odes[OdeMid],sp->Corrxs) ;
	   
        	   for (n = 0 ; n<sp->mod->n_Free ; n++){
        	       mets[n] -= v_entry(sp->Corrxs,n) ;
        	   }
        	   ModEval(sp->mod) ;
        	   ErrVal =  CalcErr(sp) ;
        	   fprintf(stderr, "err %g %d (%d)\n",ErrVal,niters, Max_Iters);
	       }
	
	       if (ErrVal <=tol)
	           *err = OK ;
	       else
		      *err = OutOfTol ;
	   }, 
	   {  /* error in meschach functions if we get here */
	       *err = SingMat ; /* most likely cause, could be other, but definite fail */
	       fprintf(stderr, "! Singular jacobian in solve/meschach !");
	   }
	) ;
	fprintf(stderr, "after %d iters returning %g  error\n",niters,ErrVal);
	       
	return (double) ErrVal ;
}			
		
			
			
			
#ifdef TEST

#include<Scampi.h>

int main(int argc,char **argv){

  ScampiModel_t m ; 
  SS_ScratchPad_t sp ; 
  void *ignore ;
  int n ;
  enum SPI_err err ;
  double tol = 1e-8 ;
  
  m = NewMD(argv[1], argv[2],&ignore) ;
  sp = NewSS_ScratchPad(m) ;
  
  ModEval(m) ;
  fprintf(stderr, "\n%d\n",m->n_Free) ;
  ModGetVec(m,Param)[0] = 1.0 ;
  ModGetVec(m,Param)[1] = 0.3 ;
  ModGetVec(m,Conc)[0] = 0.000001 ;
  ModGetVec(m,Conc)[1] = 1000.0 ;
  ModEval(m) ;
   fprintf(stderr, "\n%f\n",CalcErr(sp)) ;
  
  fprintf(stderr, "\n%f\n",Solve(sp, 10, tol, &err)) ;
  fprintf(stderr, "%g  %g\n",  ModGetVec(m,Conc)[0], ModGetVec(m,Conc)[1]) ;
  
  exit(0) ;
 }
 
 #endif
			







