
#include "../Integ_ScratchPad.h"



/* cvode Header files  */

#include "sundialstypes.h"   /* definition of type realtype                */
#include "cvode.h"           /* prototypes for CVode* functions and        */
                             /* constants CV_BDF, CV_NEWTON, CV_SV,        */
                             /* CV_NORMAL, CV_SUCCESS, and CV_ROOT_RETURN  */
#include "cvdense.h"         /* prototype for CVDense                      */
#include "nvector_serial.h"  /* definitions of type N_Vector, macro        */
                             /* NV_Ith_S, and prototypes for N_VNew_Serial */
                             /* and N_VDestroy                             */
#include "dense.h"           /* definition of type DenseMat and macro      */
                             /* DENSE_ELEM                                 */

#include "cvspgmr.h"
#include<stdlib.h>
#include<stdio.h>

#define ErrLog(err) fprintf(stderr,"ScrumPy CVODE Log: %s\n", err) ;


const double DefaultIntegTol = 1e-6 ;


struct Integ_ScratchPad {
	                                               /*** data passed to cvode functions  ***/
	realtype reltol, t, tout;                      
	N_Vector cv_y,                                 /* y data in */
			 cv_yout,                              /* y data out (can we use cv_y for both ?) */
			abstol;
	void *cvode_mem;
	int flag, flagr, iout;
	                                               /*** the model  ***/
	ScampiModel_t mod ;                           /* calculate dy from y */
	

	
} ;




void CVFunc(realtype t, N_Vector y, N_Vector ydot, void *f_data){
	/* function called by cvode to calculate new dy */
	
	Integ_ScratchPad_t sp ;
	int n ;
	double  *mets, *dmets ;
	
	sp = (Integ_ScratchPad_t) f_data ;
	mets = ModGetVec(sp->mod, Conc) ;
	

	for (n=0 ; n < sp->mod->n_Free ; n++){                         /* copy from cvode to our own vectors */
		mets[n] = (double) NV_Ith_S(y,n) ;
	}
	

   	ModEval(sp->mod) ;
	
	dmets = ModGetVec(sp->mod, dMet) ;
    for (n=0 ; n < sp->mod->n_Free ; n++){                         /* copy the result back to cvode */
        
		NV_Ith_S(ydot,n) = (realtype) dmets[n] ;
	}
	
}



Integ_ScratchPad_t NewInteg_ScratchPad(ScampiModel_t mod) {

	
	Integ_ScratchPad_t rv ;
	int n, n_Free ;
	double Tol = DefaultIntegTol, *mets ;
	
	
	rv = (Integ_ScratchPad_t) malloc(sizeof(*rv)) ;
	if (rv == NULL){
		ErrLog("\n!!\n!!Out of memory in NewInteg_ScratchPad !!\n!!");
		return NULL ;
	}

	rv->mod = mod ;                                
	n_Free = mod->n_Free ;
	mets = ModGetVec(mod,Conc) ;
	                                                   /* allocate vectors for cvode */
	rv->cv_y = N_VNew_Serial(n_Free) ;                 /* metabolite values */
	rv->cv_yout =  N_VNew_Serial(n_Free) ;             /* for values calculated by CVode() */
	rv->abstol = N_VNew_Serial(n_Free) ;               /* tolerances */
	rv->reltol = (realtype)(Tol) ;
	for (n=0 ; n<n_Free ; n++){                        /* initialise cvode vectors */
		NV_Ith_S((rv->cv_y),n) = (realtype)(mets[n]) ;
		NV_Ith_S((rv->abstol),n) = (realtype)(Tol);
	}
	rv->cvode_mem = CVodeCreate(CV_BDF, CV_NEWTON);    /* create our initial cvode "object" */ 
	CVodeMalloc(                                       /* give it it's memory  */
			rv->cvode_mem,                             /* the cvode object */
			CVFunc,                                    /* function that will calculate cv_dy from cv_y */
			*rv->mod->Time,                       /* time 0 */
			rv->cv_y,
			CV_SV,                                     /* tolerance type, copied from example - how critical  ?? */ 
			rv->reltol,
			rv->abstol
		);
	CVodeSetFdata(rv->cvode_mem, rv);                  /* tell cvode where to find the scratchpad */
	/*CVodeSetMaxNumSteps(rv->cvode_mem,2000) ;           try a bit harder */
	CVDense(rv->cvode_mem, n_Free) ;

	/* additional stability options, removed for now, seem to make things worse ! */
	
	/* CVodeSetMaxOrd(rv->cvode_mem,2) ;                                     */
	/* CVodeSetStabLimDet(rv->cvode_mem, TRUE) ;            turn on stability detection, p14 and 33 of cvode	manual */
	
	
	ErrLog("NewInteg_ScratchPad End") ;
	return rv ;
}



void ResetInteg_ScratchPad(Integ_ScratchPad_t sp){
	int n ;
	double *mets ;
	ErrLog("ResetInteg_ScratchPad\n") ;
	mets = ModGetVec(sp->mod,Conc) ;
	for (n=0 ; n<sp->mod->n_Free ; n++){                        /* initialise cvode vectors */
		NV_Ith_S((sp->cv_y),n) = (realtype) mets[n] ;
		
	}
	CVodeReInit(sp->cvode_mem, CVFunc, (realtype)(*sp->mod->Time), sp->cv_y, CV_SV,sp->reltol,sp->abstol) ;
	ErrLog("ResetInteg_ScratchPad end\n") ;

	return ;
}

void DelInteg_ScratchPad(Integ_ScratchPad_t sp){
	return ;
}


void Integrate(Integ_ScratchPad_t sp, double TimeEnd, enum SPI_err *err) {
	
	realtype tret ;
	int n ;
	double *mets ;
	
	
	fprintf(stderr, "integrate to %g",TimeEnd) ;
	
	mets = ModGetVec(sp->mod,Conc) ;
	for (n = 0 ; n< sp->mod->n_Free; n++){
		NV_Ith_S((sp->cv_y),n) = (realtype)(mets[n]) ;
	} 
	
    CVode(
		 sp->cvode_mem, 
		 (realtype) TimeEnd, 
		 sp->cv_y,
         &tret,
		 CV_NORMAL /* itask*/
	 );
	for (n = 0 ; n< sp->mod->n_Free; n++){
		mets[n] = (double) NV_Ith_S((sp->cv_y),n) ;
	} 
	fprintf(stderr, " : ended at %g\n", tret) ;
	
     return ;
}

/*
int main(){
	double mets[2] ;
	double tol = 1 ;
	int n_Free = 2 ;
	NewInteg_ScratchPad(n_Free, mets, (IntegFunc_t) NULL,tol) ;
	ErrLog("Cvode_Scratchpad ending\n") ;
	return 0 ;
}
*/
