

"""

ScrumPy -- Metabolic Modelling with Python

Copyright Mark Poolman 1995 - 2002

 This file is part of ScrumPy.

    ScrumPy is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    ScrumPy is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with ScrumPy; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

"""


import os,  sys, traceback


from Util import Set, Types

from ScrumPy.ThirdParty.Ply import yacc

import lexer
tokens = lexer.tokens
print "importing parser"

class ModelDescriptor: # intermediate representation of a model
    DicAtrs = ["StoDic", "KinDic", "OptDic","InitDic"]
    ListAtrs = ["FunList","ErrList","WarnList", "MetList", "ExtList", "IdList", "Children"]
    StringAtrs = ["FileName"]
    InitString = "import math,sys\ntry:\n    pass\n"
    CurLin = 1 # because the editor starts at 1, not 0

    def __init__(self):
        for atrs in (self.DicAtrs,dict),(self.ListAtrs,list), (self.StringAtrs,str):
            atr = atrs[1]
            for atrname in atrs[0]:
                setattr(self, atrname,atr())

        self.Parent = None

    def Merge(self,other):
        dups = Set.Intersect(self.StoDic.keys(), other.StoDic.keys())
        if len(dups) >0:
            Msg = "Reactions duplicated in " + self.FileName +  " and " + other.FileName
            Msg += "\n Human check reccomended !\n"
            self.WarnList.append((Msg,"unknown", str(dups)))  # need to identify line nos !!

        for dic in ["StoDic", "KinDic","InitDic"]:   # N.B. we ignore options in included files
            getattr(self, dic).update(getattr(other,dic))

        for lis in ["FunList", "MetList", "ExtList", "IdList"]:
            setattr(self,lis,Set.Union(getattr(self,lis),getattr(other, lis)))

    def RecordError(self, ErrorMsg, ExtraInfo=""):
        self.ErrList.append((ErrorMsg +" error :", self.CurLin, ExtraInfo))

    def RecordWarning(self, msg1="", LineNo="Unknown", msg2=""):
        self.WarnList.append((msg1, LineNo,msg2))

    def WarningsAsStr(self):
        rv = ""
        for w in self.WarnList:
            m1,line,m2 = map(str,w)
            rv += m1 + " " + m2 + " at line " + line +"\n"
        return rv

    def ReportWarnings(self, dst=sys.stdout):

        dst.write(self.WarningsAsStr())


    def PolymerCheck(self):

        Action = self.OptDic["PolymerCheck"] # it's the parsers responsibility to ensure correct values here


        for reac,sto in self.StoDic.items():

            lmets = map(lambda x:x[1],sto.lhs )
            rmets =  map(lambda x:x[1],sto.rhs )
            dupmets = []
            ldups, rdups = [],[]

            for met in lmets:
                if met in rmets:
                    dupmets.append(met)
                    ldups.append(lmets.index(met))
                    rdups.append(rmets.index(met))

            if Action == "Consumed":
                deldups = rdups
                delmets = sto.rhs
            elif Action == "Produced":
                deldups = ldups
                delmets = sto.lhs
            else:
                deldups = []

            deldups.reverse()
            for d in deldups:
                del delmets[d]

            if len(dupmets)>0:
                self.RecordWarning(reac + " produces and consumes "+",".join(dupmets) + ". Force to "+Action)

            if Action != None:
                for d in dupmets:
                    if not d in self.ExtList:
                        self.ExtList.append(d)





MDescStack = []
debug=0




##
##  ---------  Error handling functions  ------------
##

def p_error(t):  # t - token, handler for general syntax errors, invoked by yacc

    MDescStack[-1].RecordError("Syntax")
    return NextReac()

def d_error(t,msg=""): # t -token, handle errors in directive functions

    print "d_error", msg
    t = t.slice[1]

    MDescStack[-1].RecordError("Directive",msg)


def NextReac():  # scan input until start of next reaction
    tok = yacc.token()
    while tok and tok.type != "ReacId":
        tok = yacc.token()
    return tok


##
##  ------------------ helper functions  --------------------
##

def FindPairs(s,targ):

    rv = []
    pair = [-1, -1]
    for i in range(len(s)):
        if s[i] == targ:
            if  pair[0] == -1:
                pair[0] = i
            else:
                pair[1] = i
                rv.append(pair)
                pair = [-1,-1]

    if pair[1] == -1 and pair[0] != -1:
            rv.append(pair)
    return rv


def Split(St): # str.split() trips up with " delim idents contain ws, so define our own


    rv = []
    qdls = FindPairs(St,'"')

    for p in qdls:
        rv.append(St[p[0]:p[1]+1])

    for s in rv:
        St=St.replace(s,"")

    for s in St.split():
        rv.append(s)


    return rv


##
##  ---------  Directive handling functions  ------------
##




def AddToExtList(t,l):  # t,l -> token arglist

    for i in l:
        if not i in MDescStack[-1].ExtList:
            MDescStack[-1].ExtList.append(i)



def NoKinetics(t):

    MDescStack[-1].OptDic["UseKinetics"]=False


ElTypes = { # map ElType strings to python types/functions
    "rat":Types.ArbRat,
    "float":Types.Float,
    "int":Types.Int,
    "str":Types.Str,
    "bool":bool

}

def SetElType(t,l):

    if len(l) != 1:
        d_error(t,"Need exactly one argument (got " + str(len(l)) + ")")
    else:
       arg  = l[0]
       MDescStack[-1].OptDic["ElType"] = ElTypes[arg]
       lexer.MDStack[-1].IdList.remove(arg) # lexer will tread arg as an ident, which will ultimately turn up in the C code
       # this is *BAD NEWS* if arg is also a C keyword (float or int), so we remove it


def SetAutoPolymer(t,l):
    Action=l[0]
    if len(l) != 1:
        d_error(t,"Need exactly one argument (got " + str(len(l)) + ")")
    elif Action != "Consumed" and Action != "Produced":
        d_error(t, Action + " is not a valid option." )
    else:
        MDescStack[-1].OptDic["PolymerCheck"] = Action
        #lexer.MDStack[-1].IdList.remove(Action)


def SetAutoExtern(t):

    MDescStack[-1].OptDic["AutoExtern"] = True


def SetAutoDuplicateR(t):

    MDescStack[-1].OptDic["AutoDuplicateR"] = True
    if not hasattr(MDescStack[-1], "DupRs"):
        MDescStack[-1].DupRs = {}


def Include(t,l):
    for i in l:
        print i,  map(lambda md: md.FileName,  MDescStack)
        if i in map(lambda md: os.path.split(md.FileName)[-1],  MDescStack):
            MDescStack[-1].RecordError("Multiple or circular inclusion",  i)
            print "Multiple ", i
        else:
            MDescStack[-1].Children.append(Model(FName=i, Child=True, Path=MDescStack[-1].Model.Path))
            MDescStack[-1].IdList.remove(i) # remove the file name which th lexer will have treated as an identifier


def DeQuote(t):
    MDescStack[-1].OptDic["DeQuote"] = True



DirectiveActions = {
    "External":AddToExtList,
    "Structural": NoKinetics,
    "ElType":SetElType,
    "AutoExtern":SetAutoExtern,
    "AutoDuplicateR":SetAutoDuplicateR,
    "AutoPolymer":SetAutoPolymer,
    "Include":Include,
    "DeQuote":DeQuote
}


def TellMe(name,tok):
    if debug:
        print name,tok


##def p_error(t):
##
##    MDescStack[-1].RecordError("Syntax")
##
##    tok = yacc.token()
##    while tok and tok.type != "ReacId":
##        tok = yacc.token()
##    return tok
##


class Sto:
    def __init__(self, lhs, rhs, Reverse):
        self.lhs = lhs
        self.rhs = rhs
        self.Revers = Reverse

    def __repr__(self):
        rv = ""
        goes_to = self.Revers.join((" ", " "))


        for i in self.lhs:
            rv += str(i)

        rv += goes_to

        for i in self.rhs:
            rv += str(i)
        return rv



def StoTerm2MassAct(stoterm):

    if stoterm[0] == 1:
        return stoterm[1]
    else:
        return "pow(" + stoterm[1] + "," + stoterm[0] + ")"


def DefaultKinetic(Stoich):


    rv = ""
    lhs = Stoich.lhs
    rhs = Stoich.rhs

    for met in lhs[:-1]:
        rv += StoTerm2MassAct(met) + "*"
    met = lhs[-1:][0]
    rv += StoTerm2MassAct(met) + "-"

    for met in rhs[:-1]:
        rv += StoTerm2MassAct(met) + "*"
    met = rhs[-1:][0]
    rv += StoTerm2MassAct(met) +" ;"

    return rv


def RemPyModFromDic(dic):
    """remove records whose keys refer to python modules
    (only the math module at present) """


    for k in dic.keys():
        if k == "math" or k == "sys" or k.find("math.")==0 or k.find("import")==0:
            del dic[k]


def RemPyModFromLis(lis):

    for el in lis[:]:

        if el.find("math.")!=-1 or el.find("sys") != -1:
            del lis[lis.index(el)]






###################### Parser rules start here

def p_Model(t):
    """Model : Statement
            | Statement Model"""

    t[0] = t[1]


def p_Statement(t):
    """Statement : Reaction
                 | Assig
                 | Direc"""
    t[0] = t[1]


def p_Assig(t):
    """Assig : Ident Eq Expr"""
    t[0] = t[1]+t[2]+t[3]
    MDescStack[-1].InitString += "\n    " + t[0]


def p_Direc(t):
    """Direc : Directive LPar ArgList RPar
                 | Directive LPar RPar """


    if len(t.slice)==5:  # i.e.we have an arglist
        splargs = Split(t[3])
        DirectiveActions[t[1]](t,splargs)  # send t as well as the arglist so d_error(t) can be used
    else:
        try:
            DirectiveActions[t[1]](t)
        except:
            d_error(t, "Missing arguments")

    t[0] = t[1]  # no news is good news





def p_Reaction(t):
    """Reaction : ReacId  Stoich Kinetic"""

    Reac =  t[1]
    md = MDescStack[-1]
    if Reac in md.StoDic.keys():
        if not md.OptDic["AutoDuplicateR"]:
            md.RecordError("Duplicate Reaction",Reac)
        else:
            if not md.DupRs.has_key(Reac):
                md.DupRs[Reac] = 1
            else:
                md.DupRs[Reac] += 1

            sufx = "_" + str(md.DupRs[Reac])
            if Reac[-1] == '"':
                Reac = Reac[:-1]+sufx+'"'
            else:
                Reac += sufx


    md.StoDic[Reac] = t[2]

    if t[3] == "~": # DefaultKinetics
        md.KinDic[Reac] = DefaultKinetic(t[2])
    else:
        md.KinDic[Reac] = t[3]

    t[0] = Reac




def p_Stoich(t):
    """Stoich   : HalfSto Rever HalfSto
                | HalfSto Irrev HalfSto
                | HalfSto BackIrrev HalfSto"""

    t[0] = Sto(t[1],t[3],t[2])
    TellMe("Stoich",t[0])

def p_HalfSto(t):
    """ HalfSto : OneMet
                | OneMet Add HalfSto"""
    t[0] = [t[1]]
    try:
        t[0] += t[3]
    except:
        pass
    TellMe ("HalfSto",t[0])


def p_OneMet(t):
    """OneMet   : Ident
                | Int Ident
                | Real Ident
                | Rat Ident"""


    try:
        met = t[2]
        sto = t[1]
    except:
        met = t[1]
        sto = "1"

    MDesc = MDescStack[-1]
    if not met in MDesc.MetList:
        MDesc.MetList.append(met)
    t[0] = (sto,met)




def p_Kinetic(t):
    """ Kinetic : Expr
                | DefaultKin"""


    t[0]  = t[1]  # later on we will use Exprs for elasts too


precedence = (
    ('left', 'Add', 'Sub'),
    ('left', 'Mul', 'Div'),
    ('left', 'Pow')
    #('right', 'UMin')
    )


def p_Expr(t):
    """Expr : Expr Add Expr
            | Expr Sub Expr
            | Expr Mul Expr
            | Expr Div Expr
            | Power
            | UMin
            | Num
            | Fun"""

    if len(t.slice)==4:
        t[0] = t[1] + t[2] + t[3]
    else:
        t[0] = t[1]



def p_Power(t):
    """Power : Expr Pow Expr"""

    t[0] = "pow("+ t[1] + "," + t[3] + ")"

def p_UMin(t):
    """UMin : Sub Expr"""

    t[0] = t[1] + t[2]

def p_Num(t):
    """Num  : Real
            | Int
            | Rat
            | Ident"""
    t[0] = t[1]


def p_Fun(t):
    """Fun  : Ident LPar ArgList RPar
            | LPar ArgList RPar"""
    try:
        t[0] = t[1] + t[2] + t[3] + t[4]
        MDesc.FunList.append((t[1],lexer.CurLin))
    except:
        t[0] = t[1] + t[2] + t[3]




def p_ArgList(t):
    """ArgList : Expr
               | Expr Comma ArgList"""
    t[0] = t[1]
    try:
        t[0] += " " + t[3]
    except:
        pass


def GetInitErrTuple(InitD,InitString,Input):

            tb =  InitD["tb"]
            offend =  InitString.split()[tb[0][1]-1]
            lines = Input.splitlines()

            for L in lines:
                L2= L.replace(" ","")
                if L2.find(offend) != -1:
                    lineno = lines.index(L) + 1  # + 1 because editors start at line 1 rather than zero

            return lineno, offend



def DeWS(Str):
    "return copy of Str with WS removed"

    rv = ""
    for s in Str.split():
        rv += s
    return rv



def FindSrc(Input, src):  # BORKED look for src in input
    src = DeWS(src)
    Input = Input.splitlines()
    for idx in range(len(Input)):
        dei = DeWS(Input[idx])
        pos = dei.find(src)
        if pos != -1:
            return idx



parser=None

def Parse(InputStr, ModelDesc,ModelClass,Debug=False):
    "Pre: MDesc =-  ModelDiscriptor"

    if len(InputStr)==0:
        return


    MDescStack.append( ModelDesc)
    global Model
    Model = ModelClass # has to be passed in by user to avoid circular import

    lexer.Init(ModelDesc, debug=0)              # initialise the lexer

    global parser
    if parser==None:
        parser=yacc.yacc(debug=0)                        # ditto the parser
    parser.parse(InputStr)                                        # do the yaccy parse thang

    for m in ModelDesc.MetList:                                   # then add items identified as external by the "x_"/"X_" prefix
        IsExt = m[:2].upper()== "X_" or m[:3].upper() == '"X_'    # Better to do this with a parser rule ??
        if IsExt and not m in ModelDesc.ExtList:  # making sure to avoid duplicates
            ModelDesc.ExtList.append(m)

    ModelDesc.InitString += "\nexcept:\n    import traceback\n"                          # turn the init string into parseble python
    ModelDesc.InitString += "    tb = traceback.extract_tb(sys.exc_info()[2])\n"    # with somewhere to store tracebacks

    Ignore = {}
    try:
        exec ModelDesc.InitString in Ignore, ModelDesc.InitDic                # execute the python initialisations
        if "tb" in ModelDesc.InitDic.keys():   # traceback present - failed to parse InitString completely
            line, offend = GetInitErrTuple(MDesc.InitDic,MDesc.InitString,InputStr)
            ModelDesc.RecordError("Python initialisation from model", " line "+  str(line) + "- "+offend)

        else:
            RemPyModFromDic(ModelDesc.InitDic)  # remove the math module - it's not a part of the model !
            RemPyModFromLis(ModelDesc.IdList)
    except:
        print "Parser caught exception"
        # catch exec errors not caught in the tb
        report = traceback.format_exception(sys.exc_type, sys.exc_value, sys.exc_traceback)[-3:]
        # only the last 3 lines contain useful info here
        src = report[0]
        line = FindSrc(InputStr, src)
        ModelDesc.RecordError("Syntax error ", "maybe near line " + str(line) + " ("+ src+")")
    lexer.Finnish()
    MDescStack.pop()






